/* *****************************************************************************
 *  Compilation:  javac Point2D.java
 *  Execution:    java Point2D x0 y0 n
 *  Dependencies: StdDraw.java StdRandom.java
 *
 *  Immutable point data type for points in the plane.
 *
 ******************************************************************************/

import edu.princeton.cs.algs4.StdOut;
import edu.princeton.cs.algs4.StdRandom;

import java.util.Arrays;
import java.util.Comparator;

/**
 * This is an adapted version of the Algs4 Point2D class.
 * see <a href="https://algs4.cs.princeton.edu/12oop/Point2D.java.html">
 */
public class Point2D implements Comparable<Point2D> {

    private final double x;    // x coordinate
    private final double y;    // y coordinate

    /**
     * Initializes a new point (x, y).
     */
    public Point2D(double x, double y) {
        this.x = x;
        this.y = y;
    }

    /**
     * Returns the x-coordinate.
     */
    public double x() {
        return x;
    }

    /**
     * Returns the y-coordinate.
     */
    public double y() {
        return y;
    }

    /**
     * Return a string representation of this point in the format (x, y).
     */
    public String toString() {
        return "(" + x + ", " + y + ")";
    }

    /**
     * Returns the square of the Euclidean distance
     * between this point and that point.
     */
    public static double distanceSquaredTo(Point2D p, Point2D q) {
        double dx = p.x - q.x;
        double dy = p.y - q.y;
        return dx * dx + dy * dy;
    }

    /**
     * Compares by y-coordinate, breaking ties by x-coordinate.
     * Returns 0 if this point is equal to that point, 1 if this point
     * is greater than that point and -1 if that point is greater than
     * this point.
     */
    public int compareTo(Point2D that) {
        if (this.y < that.y) return -1;
        if (this.y > that.y) return +1;
        if (this.x < that.x) return -1;
        if (this.x > that.x) return +1;
        return 0;
    }

    /**
     * Returns a Comparator for comparing by distance to a reference point.
     */
    public static Comparator<Point2D> distanceToOrder(Point2D ref) {
        return new DistanceToOrder(ref);
    }

    // compare points according to their distance to a reference point.
    private static class DistanceToOrder implements Comparator<Point2D> {
        private Point2D ref; // reference point for comparison.

        private DistanceToOrder(Point2D ref) {
            this.ref = ref;
        }

        public int compare(Point2D p, Point2D q) {
            double dist1 = distanceSquaredTo(p, ref);
            double dist2 = distanceSquaredTo(q, ref);
            if (dist1 < dist2) return -1;
            else if (dist1 > dist2) return +1;
            else return 0;
        }
    }


    public static void main(String[] args) {
        int n = Integer.parseInt(args[0]);

        // Store n random points in an array.
        Point2D[] array = new Point2D[n];
        for (int i = 0; i < array.length; i++)
            array[i] = new Point2D(StdRandom.uniform(n + 1),
                    StdRandom.uniform(n + 1));

        for (Point2D p : array)
            StdOut.println(p);

        // Sort the points based on the y-coordinate, breaking ties using
        // The x-coordinate.
        Arrays.sort(array);

        StdOut.println("\nAfter Sorting: ");
        for (Point2D p : array)
            StdOut.println(p);

        // Check if the points in the array are distinct.
        boolean distinct = true;
        for (int i = 0; i < array.length - 1; i++) {
            if (array[i].compareTo(array[i + 1]) == 0) {
                distinct = false;
                break;
            }
        }
        StdOut.println("Distinct: " + distinct);

        // Sort the points based on there distance to the origin (0,0).
        Comparator<Point2D> cmp = Point2D.distanceToOrder(new Point2D(0, 0));
        Arrays.sort(array, cmp);

        StdOut.println("\nAfter Sorting with DistanceToOrder: ");
        for (Point2D p : array)
            StdOut.println(p);

        // Check if the points in the array have distinct distances from
        // the origin (in linear time in the worst case).
        distinct = true;
        for (int i = 0; i < array.length - 1; i++) {
            if (cmp.compare(array[i], array[i + 1]) == 0) {
                distinct = false;
                break;
            }
        }
        StdOut.println("Distinct: " + distinct);
    }


}
